@@ -19,7 +19,9 @@
  * 02111-1307, USA.
  */
 
+#include <config.h>
 #include "frame.h"
+#include "bell.h"
 #include "errors.h"
 #include "keybindings.h"
 
@@ -54,8 +56,11 @@ meta_window_ensure_frame (MetaWindow *window)
   frame->child_y = 0;
   frame->bottom_height = 0;
   frame->right_width = 0;
+  frame->current_cursor = 0;
 
   frame->mapped = FALSE;
+  frame->need_reapply_frame_shape = TRUE;
+  frame->is_flashing = FALSE;
   
   attrs.event_mask = EVENT_MASK;
 
@@ -123,7 +128,7 @@ meta_window_ensure_frame (MetaWindow *window)
                    window->rect.x,
                    window->rect.y);
   /* FIXME handle this error */
-  meta_error_trap_pop (window->display);
+  meta_error_trap_pop (window->display, FALSE);
   
   /* stick frame to the window */
   window->frame = frame;
@@ -142,7 +147,9 @@ meta_window_ensure_frame (MetaWindow *window)
   meta_ui_apply_frame_shape (frame->window->screen->ui,
                              frame->xwindow,
                              frame->rect.width,
-                             frame->rect.height);
+                             frame->rect.height,
+                             frame->window->has_shape);
+  frame->need_reapply_frame_shape = FALSE;
   
   meta_display_ungrab (window->display);
 }
@@ -157,6 +164,7 @@ meta_window_destroy_frame (MetaWindow *window)
 
   frame = window->frame;
   
+  meta_bell_notify_frame_destroy (frame);
   meta_ui_remove_frame (window->screen->ui, frame->xwindow);
   
   /* Unparent the client window; it may be destroyed,
@@ -179,7 +187,7 @@ meta_window_destroy_frame (MetaWindow *window)
                    /* FIXME where to put it back depends on the gravity */
                    window->frame->rect.x,
                    window->frame->rect.y);
-  meta_error_trap_pop (window->display);
+  meta_error_trap_pop (window->display, FALSE);
 
   meta_display_unregister_x_window (window->display,
                                     frame->xwindow);
@@ -253,6 +261,9 @@ meta_frame_get_flags (MetaFrame *frame)
   if (frame->window->fullscreen)
     flags |= META_FRAME_FULLSCREEN;
 
+  if (frame->is_flashing)
+    flags |= META_FRAME_IS_FLASHING;
+
   return flags;
 }
 
@@ -275,6 +286,20 @@ meta_frame_calc_geometry (MetaFrame         *frame,
   *geomp = geom;
 }
 
+static void
+update_shape (MetaFrame *frame)
+{
+  if (frame->need_reapply_frame_shape)
+    {
+      meta_ui_apply_frame_shape (frame->window->screen->ui,
+                                 frame->xwindow,
+                                 frame->rect.width,
+                                 frame->rect.height,
+                                 frame->window->has_shape);
+      frame->need_reapply_frame_shape = FALSE;
+    }
+}
+
 void
 meta_frame_sync_to_window (MetaFrame *frame,
                            int        resize_gravity,
@@ -282,8 +307,11 @@ meta_frame_sync_to_window (MetaFrame *frame,
                            gboolean   need_resize)
 {
   if (!(need_move || need_resize))
-    return;
-  
+    {
+      update_shape (frame);
+      return;
+    }
+
   meta_topic (META_DEBUG_GEOMETRY,
               "Syncing frame geometry %d,%d %dx%d (SE: %d,%d)\n",
               frame->rect.x, frame->rect.y,
@@ -299,18 +327,17 @@ meta_frame_sync_to_window (MetaFrame *frame,
                                   frame->rect.width,
                                   frame->rect.height);
 
-      /* Done before the window resize, because doing it before means
-       * part of the window being resized becomes unshaped, which may
-       * be sort of hard to see with bg = None. If we did it after
-       * window resize, part of the window being resized would become
-       * shaped, which might be more visible.
-       */
-      
-      meta_ui_apply_frame_shape (frame->window->screen->ui,
-                                 frame->xwindow,
-                                 frame->rect.width,
-                                 frame->rect.height);
+      /* we need new shape if we're resized */
+      frame->need_reapply_frame_shape = TRUE;
     }
+
+  /* Done before the window resize, because doing it before means
+   * part of the window being resized becomes unshaped, which may
+   * be sort of hard to see with bg = None. If we did it after
+   * window resize, part of the window being resized would become
+   * shaped, which might be more visible.
+   */
+  update_shape (frame);
   
   if (need_move && need_resize)
     XMoveResizeWindow (frame->window->display->xdisplay,
@@ -351,3 +378,21 @@ meta_frame_queue_draw (MetaFrame *frame)
   meta_ui_queue_frame_draw (frame->window->screen->ui,
                             frame->xwindow);
 }
+
+void meta_frame_set_screen_cursor (MetaFrame	*frame,
+				   MetaCursor	cursor)
+{
+  Cursor xcursor;
+  if (cursor == frame->current_cursor)
+    return;
+  frame->current_cursor = cursor;
+  if (cursor == META_CURSOR_DEFAULT)
+    XUndefineCursor (frame->window->display->xdisplay, frame->xwindow);
+  else
+    { 
+      xcursor = meta_display_create_x_cursor (frame->window->display, cursor);
+      XDefineCursor (frame->window->display->xdisplay, frame->xwindow, xcursor);
+      XFreeCursor (frame->window->display->xdisplay, xcursor);
+    }
+}
+
