@@ -48,6 +48,14 @@ typedef enum
   META_SCREEN_BOTTOMRIGHT
 } MetaScreenCorner;
 
+typedef enum
+{
+  META_SCREEN_UP,
+  META_SCREEN_DOWN,
+  META_SCREEN_LEFT,
+  META_SCREEN_RIGHT
+} MetaScreenDirection;
+
 struct _MetaScreen
 {
   MetaDisplay *display;
@@ -70,6 +78,8 @@ struct _MetaScreen
 
   MetaCursor current_cursor;
 
+  Window flash_window;
+
   Window wm_sn_selection_window;
   Atom wm_sn_atom;
   Time wm_sn_timestamp;
@@ -124,12 +134,30 @@ void          meta_screen_ensure_tab_popup    (MetaScreen                 *scree
 
 void          meta_screen_ensure_workspace_popup (MetaScreen *screen);
 
-void          meta_screen_focus_top_window    (MetaScreen                 *screen,
-                                               MetaWindow                 *not_this_one);
+void          meta_screen_focus_top_window     (MetaScreen                 *screen,
+                                                MetaWindow                 *not_this_one);
+void          meta_screen_focus_mouse_window   (MetaScreen                 *screen,
+                                                MetaWindow                 *not_this_one);
+void          meta_screen_focus_default_window (MetaScreen                 *screen,
+                                                MetaWindow                 *not_this_one);
 
-const MetaXineramaScreenInfo* meta_screen_get_current_xinerama (MetaScreen *screen);
-const MetaXineramaScreenInfo* meta_screen_get_xinerama_for_window (MetaScreen *screen,
-								   MetaWindow *window);
+
+const MetaXineramaScreenInfo* meta_screen_get_current_xinerama    (MetaScreen    *screen);
+const MetaXineramaScreenInfo* meta_screen_get_xinerama_for_rect   (MetaScreen    *screen,
+                                                                   MetaRectangle *rect);
+const MetaXineramaScreenInfo* meta_screen_get_xinerama_for_window (MetaScreen    *screen,
+                                                                   MetaWindow    *window);
+
+gboolean      meta_screen_window_intersects_xinerama (MetaScreen *screen, 
+                                                      MetaWindow *window,
+                                                      int         which_xinerama);
+
+const MetaXineramaScreenInfo* meta_screen_get_xinerama_neighbor (MetaScreen *screen,
+                                                                 int         which_xinerama,
+                                                                 MetaScreenDirection dir);
+void          meta_screen_get_natural_xinerama_list (MetaScreen *screen,
+                                                     int**       xineramas_list,
+                                                     int*        n_xineramas);
 
 void          meta_screen_update_workspace_layout (MetaScreen             *screen);
 void          meta_screen_update_workspace_names  (MetaScreen             *screen);
@@ -138,10 +166,23 @@ void          meta_screen_queue_workarea_recalc   (MetaScreen             *scree
 Window meta_create_offscreen_window (Display *xdisplay,
                                      Window   parent);
 
-void meta_screen_calc_workspace_layout (MetaScreen *screen,
-                                        int         num_workspaces,
-                                        int        *r,
-                                        int        *c);
+typedef struct MetaWorkspaceLayout MetaWorkspaceLayout;
+
+struct MetaWorkspaceLayout
+{
+  int rows;
+  int cols;
+  int *grid;
+  int grid_area;
+  int current_row;
+  int current_col;
+};
+
+void meta_screen_calc_workspace_layout (MetaScreen          *screen,
+                                        int                  num_workspaces,
+                                        int                  current_space,
+                                        MetaWorkspaceLayout *layout);
+void meta_screen_free_workspace_layout (MetaWorkspaceLayout *layout);
 
 void meta_screen_resize (MetaScreen *screen,
                          int         width,
@@ -152,5 +193,8 @@ void meta_screen_resize (MetaScreen *screen,
 void     meta_screen_show_desktop   (MetaScreen *screen);
 void     meta_screen_unshow_desktop (MetaScreen *screen);
 
+void     meta_screen_apply_startup_properties (MetaScreen *screen,
+                                               MetaWindow *window);
+
 
 #endif
